/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz UK LLP
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 **/

var snmpGet = function (message, resourceLocator) {

    var snmp = resourceLocator.netSnmp;
    var getNative = resourceLocator.snmpGetNative(JSON.parse(JSON.stringify(message)), resourceLocator);
    var cmdUtils = resourceLocator.utils.command;
    var snmpCommon = resourceLocator.snmpCommon;

    var cid = message.correlationId;
    var options = message.payload;

    var oidsMapping = cmdUtils.extractFromOptions(options, 'oids', {});
    var host = cmdUtils.extractFromOptions(options, 'host', 'localhost');
    var community = cmdUtils.extractFromOptions(options, 'community', 'public');

    var oids = snmpCommon.extractOids(oidsMapping);
    var transformer = message.customTransformer ? message.customTransformer : snmpCommon.transformer;

    // Comparison to null is mandatory because snmp.Version1 = 0
    if (options.version == null) { //jshint ignore:line
        options.version = snmp.Version2c;
    }

    function extractOidsResponse(response) {
        console.info("SNMP_GET - (%s): Got response - number of returned oids: %s", cid, response.length);
        var values = {};
        for (var i = 0; i < response.length; i++) {
            var oid = response[i].oid;

            if (!(snmp.isVarbindError(response[i]))) {
                var value = transformer(oidsMapping[oid])(response[i].value);
                console.debug("SNMP_GET - (%s): - SNMP value retrieved: %s=%s", cid, oid, value);
                values[oid] = value;
            }
            else {
                console.warn("SNMP_GET - (%s): SNMP get - error retrieving oid: %s => varbind error: s",
                    cid, oid, snmp.varbindError(response[i]));
            }
        }
        return {"oids": values};
    };

    function extractErrorResponse(error, response) {
        if (snmpCommon.isEmpty(error) && response === undefined) {
            console.info("SNMP_GET - (%s): warn: ignoring callback since error is empty and response is undefined", cid);
            response = {"error": error};
        } else {
            console.warn("SNMP_GET - (%s): SNMP error received: %s", cid, error.toString());
            response = {"error": error};
        }
        return response;
    };

    return {
        execute: function (callback) {

            console.debug('SNMP_GET - (%s): Sending SNMP get request to host %s for oids: %s', cid, host, JSON.stringify(oids));

            var snmpSession = snmp.createSession(host, community, options);

            snmpSession.on("error", function (error) {
                var errorMessage = error.toString();
                console.error("- SNMP Get - (socket error) %s", errorMessage);
                if (error.name === 'RangeError') {
                    getNative.execute(callback);
                } else {
                    try {
                        snmpSession.close();
                    } catch (error) {
                        console.error("SNMP_GET - cannot close session: " + error);
                    } finally {
                        callback({'error': errorMessage});
                    }
                }
            });

            snmpSession.get(oids, function (error, response) {
                if (error) {
                    response = extractErrorResponse(error, response);
                } else if (response !== null && response !== undefined) {
                    response = extractOidsResponse(response);
                }
                else {
                    response = {'error': "invalid SNMP response"};
                }
                console.debug("SNMP_GET - (%s): command execution concluded, yielding result: %s", cid, JSON.stringify(response));
                try {
                    snmpSession.close();
                } catch (error) {
                    console.error("SNMP_GET - cannot close session: " + error);
                } finally {
                    callback(response);
                }
            });
        },
        describe: function () {
            return cid + " - snmpGet - " + host + ' - ' + JSON.stringify(oids);
        }
    };

};

module.exports.command = snmpGet;