/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz UK LLP
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Listens for broadcast DHCP lease packets and sends the data to the backend
 *
 * Created by Iacopo Papalini <iacopo@domotz.com>
 *
 * 35 seconds: the assumption is that in the meantime the fast discovery
 * has produced the creation of the missing device
 */
const RETRY_INTERVAL = 35000;
const LOG_PREFIX = 'DHCP_LISTEN - ';


var dhcpDiscoveryFactory = function (dhcpjs, random, engineRequest, resourceLocator) {
    "use strict";

    var my_console = resourceLocator.log.decorateLogs(LOG_PREFIX);

    function extractMacIfWellFormed(message) {
        var error;
        if (!message.op) {
            error = "Corrupted package, no 'op' field.";
            my_console.verbose(error);
            return [error];
        }
        if (message.op.value !== 1) {
            error = "Not a BOOTREQUEST. Value: " + message.op.value;
            my_console.verbose(error);
            return [error];
        }
        if (!message.options) {
            error = "Not a DHCP request";
            my_console.verbose(error);
            return [error];
        }
        if (!message.options.dhcpMessageType || message.options.dhcpMessageType.name !== 'DHCPDISCOVER') {
            error = "Not a DHCPDISCOVER packet, ignoring";
            my_console.verbose(error);
            return [error];
        }

        var hwaddr = message.options.clientIdentifier || message.options.chaddr;

        if (!hwaddr || !hwaddr.type || hwaddr.type.value !== 1) {
            error = "Not an ethernet device";
            my_console.verbose(error);
            return [error];
        }

        return [null, hwaddr.address.toUpperCase()];
    }

    function signalToEngine(mac, dhcpData) {
        my_console.info("Sending DHCP fingerprint for device mac: %s) ", mac);
        return engineRequest.cachedSendWithPromise('/subnet/ip/device/' + mac + '/inspection-data/dhcp', 'PUT', dhcpData, null, null, {
            maxAttempts: 1,
            retryDelay: RETRY_INTERVAL
        });
    }

    function handleMessage(message) {
        var res = extractMacIfWellFormed(message);
        var error = res[0];
        if (error) {
            return error;
        }
        var mac = res[1];
        if (mac !== false) {
            my_console.verbose("DHCP Packet arrived: mac=%s", mac);
            signalToEngine(mac, message.options);
        } else {
            my_console.verbose("Ignoring DHCP message");
        }
    }

    function onListening(address) {
        my_console.info('- listening for DHCP packets on ' + address);
    }

    function listenForLeasePackets() {
        var server = dhcpjs.createServer(null, {type: 'udp4', reuseAddr: true});
        server.on('listening', onListening);
        server.on('message', handleMessage);
        server.bind();
    }

    return {
        // ----- TEST -----
        _handleMessage: handleMessage,
        // ----- TEST -----
        listenForLeasePackets: listenForLeasePackets
    };
};

module.exports.factory = dhcpDiscoveryFactory;