/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 **/
/**
 * Created by Tommaso Latini <tommaso@domotz.com> on 25/07/15.
 *
 * This module recover host information on a UNIX-like operating system (like
 * Debian, DSM, Qnap OS).
 *
 * These information are fundamental for the application, so we recover it
 * synchronously at boot and we exit in case of failure.
 * The periodical refresh is asynchronous for not blocking nodejs mainloop.
 */

const MODULE_NAME = 'HOST_INTERFACES';
const IP_OCTET = "([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])";
const IP_VALIDATOR = new RegExp("^(" + IP_OCTET + "\.){3}" + IP_OCTET + "$");
const LINK_LOCAL = new RegExp("^169\.254\." + IP_OCTET + "\." + IP_OCTET + "$");
const MAC_VALIDATOR = /^(?!(?:ff:ff:ff:ff:ff:ff|00:00:00:00:00:00))(?:[\da-f]{2}:){5}[\da-f]{2}$/i;

function logInterface(name, mac, ip, netmask) {
    console.info('%s - (%s): MAC address: [%s] - ' +
        'IP address: [%s] - Netmask: [%s].', MODULE_NAME, name, mac,
        ip, netmask);
}

function checkIp(addr) {
    if (!IP_VALIDATOR.test(addr) || LINK_LOCAL.test(addr)) {
        throw new Error(MODULE_NAME + " - Invalid IP address. [" + addr + "] " +
            "Maybe the interface is not connected.");
    }
}

function checkMac(mac) {
    if (!mac || !MAC_VALIDATOR.test(mac.toLowerCase())) {
        throw new Error(MODULE_NAME + " - Invalid MAC address [" + mac + "]. " +
            "Maybe not a valid interface!");
    }
}

var factory = function (proc, ipUtils, networkInterfaces, customInterfaces, nmapXml2Js, lodash, q, dns, fs) {

    /**
     * [**interfacesList] should be an hash of couple:
     *    <interface_name>: { mac: <mac_address> }
     * At least mac address MUST be into interfaces list because
     * it is used by licence api for binding domotz agent.
     *
     * [**defaultInterface] is the name of the interface of the default gateway
     */
    var interfacesList = {},
        defaultGateway = null,
        defaultGatewayMAC = null,
        defaultInterface = null,
        fullIpList = [];

    var localNetworkPromise = q.defer();
    var localNetworkInfoAvailable = localNetworkPromise.promise;

    var recoverHostInfo = networkInterfaces ? recoverHostInfoV1 : recoverHostInfoV2;

    function mergeCustomInterfaces(tmp) {
        var previousCustomInterfaces = lodash.omitBy(interfacesList, function (value) {
            return value.custom != true;
        });
        interfacesList = lodash.extend(tmp, previousCustomInterfaces);
    }

    function recoverHostInfoV1() {
        return proc.spawn('ip', ['route'])
            .then(function (data) {
                var lines = data.toString().split(/\r?\n/);
                lines.every(function (item) {
                    var fields = item.split(/ +/);
                    console.debug(MODULE_NAME + ': route line = > %s', item, {});
                    if (fields[0] === "default" && fields[1] === "via") {
                        console.debug(MODULE_NAME + ': Fields = %j', fields, {});
                        defaultGateway = fields[2];
                        console.info(MODULE_NAME + ': Detected gateway => %s', defaultGateway, {});
                        return false;
                    }
                    return true;
                });
            })
            .then(networkInterfaces.getDefaultGatewayAsync)
            .then(function (descr) {
                defaultInterface = descr.interface;
            })
            .then(networkInterfaces.getInterfacesAsync)
            .then(function (list) {
                if (!list || !list.length) {
                    throw new Error(MODULE_NAME + ': No valid interfaces in the final list');
                }
                var tmp = {};
                list.forEach(function (nic) {
                    var name = nic.name;
                    logInterface(name, nic.mac, nic.ip, nic.netmask, defaultInterface);
                    delete nic.name;
                    tmp[name] = nic;
                });
                mergeCustomInterfaces(tmp);
            })
            .then(getGatewayMac);
    }


    function recoverHostInfoV2() {
        var path = require('path');
        return proc.spawn(path.join(process.env.DOMOTZ_BIN_DIR, 'domotz_network_utils'))
            .then(function (output) {
                var rows = output.split('\n');
                var tmp = {};
                for (var i = 0; i < rows.length && rows[i].length; i++) {
                    if (rows[i] === '---') {
                        // ---
                        // Default Gateway: 192.168.77.1 (or empty)
                        // Interface: eth1
                        // IPv4: 192.168.77.5
                        // MAC: 08:00:27:53:3a:53
                        // CIDR: 24
                        var defaultGatewayFound = false;
                        if (!defaultGateway) {
                            defaultGateway = rows[++i].slice(17);
                            defaultGatewayFound = true;
                        } else {
                            i++;
                        }
                        var name = rows[++i].slice(11);
                        var ip = rows[++i].slice(6);
                        var mac = rows[++i].slice(5).toUpperCase();
                        var netmask = parseInt(rows[++i].slice(6));
                        try {
                            checkIp(ip);
                            checkMac(mac);
                            if (defaultGateway && defaultGatewayFound) {
                                defaultInterface = name;
                                console.info('%s - (%s): Default Gateway on %s',
                                    MODULE_NAME, name, defaultGateway);
                            }
                            logInterface(name, mac, ip, netmask);
                            tmp[name] = {
                                ip: ip,
                                mac: mac,
                                netmask: netmask
                            };
                        } catch (err) {
                            console.warn(err.message);
                        }
                    }
                }
                mergeCustomInterfaces(tmp);
            })
            .then(getGatewayMac)
            .then(function () {
                localNetworkPromise.resolve();
            });
    }

    function getGatewayMac() {
        "use strict";
        console.info("%s - Searching for Gateway MAC address", MODULE_NAME);
        if (!defaultGateway) {
            console.info("%s - No Default Gateway MAC", MODULE_NAME);
            return;
        }
        return proc.spawn('domotz_nmap', ['-sP', '-oX', '-', defaultGateway]).then(
            function (output) {
                console.debug("%s - Parsing: %s", MODULE_NAME, output);
                nmapXml2Js.toJSON(output).then(
                    function (xml) {
                        console.debug("%s - %s", MODULE_NAME, JSON.stringify(xml));
                        xml.nmaprun.host[0].address.every(function (addr) {
                            if (addr.$.addrtype === 'mac') {
                                defaultGatewayMAC = addr.$.addr;
                                console.info("%s - Gateway MAC address: %s", MODULE_NAME, defaultGatewayMAC);
                                return false;
                            }
                            return true;
                        });

                    }
                ).catch(function (e) {
                    console.warn("%s - Cannot parse arp output to get default gateway MAC, using onw MAC: %s", MODULE_NAME, output);
                    defaultGatewayMAC = interfacesList[0].mac;
                });

            }
        ).catch(function (e) {
            console.error("%s Error getting default gateway %s", MODULE_NAME, JSON.stringify(e));
        });
    }

    function refreshCustomInterfaces(customSubnets) {
        var networkInterfaces = lodash.omitBy(interfacesList, function (element) {
            return element.custom === true;
        });
        var subnets;
        if (!customSubnets || Object.keys(customSubnets).length === 0) {
            //try to load legacy file-based interfaces. Returns empty object as a fallback
            subnets = customInterfaces.get();
        } else {
            subnets = customSubnets;
        }
        interfacesList = lodash.extend(subnets, networkInterfaces);
    }

    recoverHostInfo();
    refreshCustomInterfaces();

    var getCustomInterfaces = function () {
        return lodash.omitBy(interfacesList, function (el) {
            return el.custom !== true;
        });
    };

    function commonChecks(device) {
        var customInterfaces = getCustomInterfaces();
        if (!customInterfaces)
            return false;
        if (Object.keys(customInterfaces).length === 0) {
            return false;
        }
        return device.length;
    }

    var deviceBelongsToExternalSubnet = function (device) {
        if (!commonChecks(device)) {
            return false;
        }

        for (var k = 0; k < device.length; k++) {
            if (isInCustomSubnet((device[k]['ip']))) {
                return true;
            }
        }
        return false;
    };

    var deviceBelongsToExternalHosts = function (device) {
        if (!commonChecks(device)) {
            return false;
        }

        for (var k = 0; k < device.length; k++) {
            if (getExternalHostMac((device[k]['ip']))) {
                return true;
            }
        }
        return false;
    };

    function getExternalHostMac(ipAddr) {
        var customInterfaces = getCustomInterfaces();
        for (var customId in customInterfaces) {
            var custom = customInterfaces[customId];
            if (custom.discovery_hw_address && custom.ip == ipAddr) {
                return custom.discovery_hw_address;
            }
        }
        return false;
    }

    function isInCustomSubnet(ipAddr) {
        var customInterfaces = getCustomInterfaces();
        for (var customId in customInterfaces) {

            var custom = customInterfaces[customId];
            if (!custom.ip) {
                continue;
            }
            var netAddress = new ipUtils.Address4(custom.ip + '/' + custom.netmask);
            if (custom.netmask == 32 && ipAddr === custom.ip) {
                return true;
            }
            var addr = new ipUtils.Address4(ipAddr);
            if (addr.isInSubnet(netAddress)) {
                return true;
            }
        }
        return false;
    }

    function resolveExternalHosts() {
        console.info("%s - Resolving external hosts ...", MODULE_NAME);
        console.info("%s - %s", MODULE_NAME, JSON.stringify(interfacesList));

        var promises = [];

        for (var name in interfacesList) {
            var i = interfacesList[name];
            if (i.custom && i.hostname) {
                promises.push(function (obj) {
                    var p = q.defer();
                    console.info("%s - Resolving ", MODULE_NAME, JSON.stringify(obj.hostname));
                    dns.lookup(obj.hostname, function (err, ip) {
                        if (err) {
                            console.warn("%s - Error resolving hostname %s: %s", MODULE_NAME, obj.hostname, JSON.stringify(err));
                            p.resolve();
                            return;
                        }
                        console.debug("%s - Successfully resolved hostname %s: %s", MODULE_NAME, obj.hostname, JSON.stringify(ip));
                        if (typeof ip === 'string') {
                            obj.ip = ip;
                        } else {
                            obj.ip = ip[0];
                        }
                        p.resolve();
                    });
                    return p.promise;
                }(i))
            }
        }
        return q.all(promises);
    }

    function setFullIpList(l) {
        fullIpList = lodash.uniq(l);
    }


    return {
        recoverHostInfo: recoverHostInfo,
        getIP: function getIP(nic) {
            return interfacesList[nic || defaultInterface].ip;
        },
        getMAC: function (nic) {
            return interfacesList[nic || defaultInterface].mac;
        },
        getNetmask: function (nic) {
            return interfacesList[nic || defaultInterface].netmask;
        },
        getDefaultGateway: function () {
            return defaultGateway;
        },
        getDefaultGatewayMAC: function () {
            return defaultGatewayMAC;
        },
        getInterfacesList: function () {
            lodash.forOwn(interfacesList, function (value, key) {
                interfacesList[key]['ipList'] = [];
            });

            lodash.forEach(fullIpList, function (raw_ip) {
                var ip = new ipUtils.Address4(raw_ip);
                lodash.forOwn(interfacesList, function (nic) {
                    if (nic.ip) {
                        var nicIp = new ipUtils.Address4(nic.ip + '/' + nic.netmask);
                        if (ip.isInSubnet(nicIp)) {
                            nic.ipList.push(raw_ip)
                        }
                    }
                });
            });

            return interfacesList;
        },
        isCustomInterface: function (nic) {
            var iface = interfacesList[nic];
            return iface.custom && iface.custom === true;
        },
        isExternalHostInterface: function (nic) {
            var iface = interfacesList[nic];
            if (iface.custom && iface.custom === true && iface.discovery_hw_address)
                return true;
            return false
        },
        getInterface: function () {
            return defaultInterface;
        },
        getInterfaceType: function (ip) {
            for (var iface in interfacesList) {
                if (interfacesList.hasOwnProperty(iface)) {
                    var nic = interfacesList[iface];
                    var nicIp = new ipUtils.Address4(nic.ip + '/' + nic.netmask);
                    var secondIp = new ipUtils.Address4(ip);
                    if (secondIp.isInSubnet(nicIp)) {
                        return nic.custom || false;
                    }
                }
            }
            return false;
        },
        loadCustomInterfaces: refreshCustomInterfaces,
        deviceBelongsToExternalSubnet: deviceBelongsToExternalSubnet,
        deviceBelongsToExternalHosts: deviceBelongsToExternalHosts,
        getExternalHostMac: getExternalHostMac,
        resolveExternalHosts: resolveExternalHosts,
        IPisInCustomSubnet: isInCustomSubnet,
        setFullIpList: setFullIpList,
        localNetworkInfoAvailable: localNetworkInfoAvailable
    };
};

module.exports.factory = factory;