/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * This module has the responsibility to coordinate both the delta and the full IP devices
 * discovery.
 * It calls the InterfacesBindingStorage to update the interfaces binding discovered by the
 * BindingDiscoverer. Then it sends the results to the engine.
 *
 * Created by Iacopo Papalini <iacopo@domotz.com> on 18/05/16.
 */



module.exports.factory = function (event,
                                   bindingDiscoverer,
                                   interfacesBinding,
                                   engineReportingStrategy,
                                   discoveryRetry,
                                   host,
                                   log) {
    "use strict";

    var performing = false;
    var iteration = 0;

    var my_console = log.decorateLogs('IP_DISCOVERY - ');

    function performDiscovery() {
        if (performing) {
            my_console.warn("(%d): Avoiding double run of IP network devices discovery", iteration);
            return;
        }
        iteration++;
        performing = true;

        my_console.verbose("(%d): Performing IP network devices discovery", iteration);

        return interfacesBinding.assureInitialized().
            then(interfacesBinding.staleAll).
            then(setIteration).
            then(host.resolveExternalHosts).
            then(bindingDiscoverer.discoverAll).
            then(interfacesBinding.setFreshDiscoveredBindings).
            then(discoveryRetry.retryOnStaleBindingsLocalDevices).
            then(discoveryRetry.retryOnStaleBindingsExternalDevices).
            then(engineReportingStrategy.reportDiscovery).
            then(interfacesBinding.purgeStale).
            then(event.endFastDeviceDiscovery).
            catch(errorHandler).
            finally(function () {
                performing = false;
            });
    }

    function setIteration() {
        bindingDiscoverer.setIteration(iteration);
    }

    function errorHandler(err) {
        my_console.error("(%d): Discovery error: %s", iteration, err.message);
        my_console.error(err.stack);
    }

    return {
        performDiscovery: performDiscovery
    };
};
