/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz UK LLP
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 **/
/**
 *
 * Created by Massimiliano Cuzzoli <massi@domotz.com> on 05/10/16.
 *
 */


const PROTOCOL_NAME = 'snmp';
const DISCOVERY_SEMAPHORE = 'SNMP_SEMAPHORE';
const OIDS_TO_DISCOVER = {
    '1.3.6.1.2.1.1.1.0': 'string',  // sysDescr
    '1.3.6.1.2.1.1.2.0': 'string',  // sysOID
// IGNORED  ---  '1.3.6.1.2.1.1.3': 'int',     // sysUpTimeInstance
    '1.3.6.1.2.1.1.4.0': 'string',  // sysContact
    '1.3.6.1.2.1.1.5.0': 'string',  // sysName
    '1.3.6.1.2.1.1.6.0': 'string',  // sysLocation
    '1.3.6.1.2.1.1.7.0': 'int',     // sysServices
    '1.3.6.1.2.1.2.1.0': 'int'      // ifNumber
};
const DEFAULT_READ_COMMUNITY = 'public';
const OID_PORT_NUMBER = '1.3.6.1.2.1.2.1.0';


var snmpDiscoveryFactory = function (resourceLocator, commandBuilder, responseRendering, snmpGetterFactory) {
    "use strict";
    var my_console = resourceLocator.log.decorateLogs("SNMP DISCOVERY - ");
    var snmpInterfacesNumberStorage = resourceLocator.snmpInterfacesNumberStorage;
    var communitiesProvider = resourceLocator.snmpCommunitiesProvider;
    var interfacesBinding = resourceLocator.interfacesBindingStorage;
    var mutex = resourceLocator.utils.mutex;
    var q = resourceLocator.q;
    var snmpDiscoverySnapshotData = {};

    var deferredSnmpBulkGet = function (snmpCommand, host, mac) {
        var deferred = q.defer();

        snmpGetterFactory(snmpCommand, resourceLocator).execute(
            function (result) {
                my_console.debug('Received result for %s -> %s', mac, JSON.stringify(result));
                result.host = host;
                result.mac = mac;
                deferred.resolve(result);
            }
        );

        return deferred.promise;
    };

    var lock_discovery = function () {
        return mutex.lock(DISCOVERY_SEMAPHORE, 1);
    };

    var unlock_discovery = function () {
        mutex.unlock(DISCOVERY_SEMAPHORE, 1);
    };

    var renderSnmpResults = function (results) {
        for (var i = 0; i < results.length; i++) {
            var host = results[i].host;
            var mac = results[i].mac;
            var data = null;
            try {
                data = responseRendering(resourceLocator, results[i]);
            } catch (e) {
                data = undefined;
            }
            if (host === undefined || mac === undefined || data === null) {
                my_console.error("Something went wrong in rendering the SNMP result, skipping");

            } else {
                my_console.debug('Parsed Result: %s - |%s|', mac, JSON.stringify(data));

                if (!data.error) {
                    snmpDiscoverySnapshotData[mac] = data;
                    if (data.oids && data.oids[OID_PORT_NUMBER]) {
                        var number = parseInt(data.oids[OID_PORT_NUMBER], 10);
                        snmpInterfacesNumberStorage.store(mac, number);
                    }
                }
            }
        }
    };

    var discover = function (callback, err_callback) {
        if (!lock_discovery()) {
            my_console.warn("WARNING: SNMP Discovery already in progress, ignoring the call");
            return q(null);
        }

        return communitiesProvider.get().then(function (communities) {
            var deferred_result_list = [];
            interfacesBinding.listFresh().forEach(function (device) {
                var community = (communities[device.mac] && communities[device.mac].snmp_read_community) || DEFAULT_READ_COMMUNITY;
                if (community !== DEFAULT_READ_COMMUNITY) {
                    my_console.debug("Using custom read community: %s", community);
                }
                var address = device.ip;
                var snmpCommand = commandBuilder.buildBulkGetCommand(OIDS_TO_DISCOVER, address, community);
                my_console.debug("Starting SNMP discovery for %s [mac %s]", address, device.mac);
                deferred_result_list.push(deferredSnmpBulkGet(snmpCommand, address, device.mac));
            });
            return q.all(deferred_result_list)
                .then(renderSnmpResults)
                .then(function () {
                    unlock_discovery();
                    if (callback) {
                        callback(snmpDiscoverySnapshotData);
                    }
                });
        }).catch(function (err, err2) {
            unlock_discovery();
            my_console.error("Error on snmp information processing %s %s", err, err2);
            if (err_callback) {
                err_callback(err);
            }
        });
    };

    var applyFunctionToSnapshotEntries = function (functionToApply) {
        my_console.debug("post-processing: " + JSON.stringify(snmpDiscoverySnapshotData));
        for (var mac_address in snmpDiscoverySnapshotData) {
            if (snmpDiscoverySnapshotData.hasOwnProperty(mac_address)) {
                functionToApply(mac_address, snmpDiscoverySnapshotData[mac_address], PROTOCOL_NAME);
            }
        }
    };

    var searchEntryIntoSnapshot = function (device_address) {
        return snmpDiscoverySnapshotData[device_address];
    };

    return {
        discover: discover,
        applyFunctionToSnapshotEntries: applyFunctionToSnapshotEntries,
        searchEntryIntoSnapshot: searchEntryIntoSnapshot,
        isDiscoveryInProgress: function () {
            return mutex.isLocked(DISCOVERY_SEMAPHORE);
        },
        getProtocolName: function () {
            return PROTOCOL_NAME;
        },
        getProtocolNameUpperCase: function () {
            return PROTOCOL_NAME.toUpperCase();
        },
        getOidForInterfaceNumber: function () {
            return OID_PORT_NUMBER;
        }
    };
};


module.exports.factory = snmpDiscoveryFactory;