/** This file is part of Domotz Agent.
 * Copyright (C) 2017  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * Created by Alessio Sanfratello <asanfratello@domotz.com> on 09/02/17.
 */

const DISCOVERY_SEMAPHORE = 'SNMP_CUSTOM_SEMAPHORE';
const DEFAULT_READ_COMMUNITY = 'public';
const SNMP_DISCOVERY_TIMEOUT = 10000;

var snmpDiscoveryFactory = function (resourceLocator, commandBuilder, responseRendering, snmpGetterFactory) {
    "use strict";

    var mutex = resourceLocator.utils.mutex;
    var q = resourceLocator.q;
    var engineRequest = resourceLocator.engineRequest;
    var snmpCommon = resourceLocator.snmpCommon;
    var snmpDiscoverySnapshotData = {};
    var macToOidList = null;
    var my_console = resourceLocator.log.decorateLogs("SNMP CUSTOM DISCOVERY - ");
    var interfacesBinding = resourceLocator.interfacesBindingStorage;

    function invalidate() {
        macToOidList = null;
    }

    var lock_discovery = function () {
        return mutex.lock(DISCOVERY_SEMAPHORE, 1);
    };

    var unlock_discovery = function () {
        mutex.unlock(DISCOVERY_SEMAPHORE, 1);
    };

    /**
     * Get values from oid sensors
     * @param snmpCommand oid resource (Ex: 1.2.1.2)
     * @param host device ip address
     * @param mac mac address of the device
     */

    var deferredSnmpBulkGet = function (snmpCommand, host, mac) {
        var deferred = q.defer();

        snmpGetterFactory(snmpCommand, resourceLocator).execute(
            function (result) {
                my_console.debug('Received result for %s: %s Bytes', host, JSON.stringify(result).length);
                result.host = host;
                result.mac = mac;
                deferred.resolve(result);
            }
        );

        return deferred.promise;
    };


    var renderDiscoveryData = function (rawResultList) {
        my_console.debug('Rendering discover raw list : |%s|', JSON.stringify(rawResultList));
        rawResultList.forEach(function (rawResult) {
            var ipAddress = rawResult.host;
            var mac = rawResult.mac;
            var data = responseRendering(resourceLocator, rawResult);

            my_console.debug('Parsed Result: %s - |%s|', ipAddress, JSON.stringify(data));

            if (data.error === false) {
                snmpDiscoverySnapshotData[mac] = data.oids;
            }
        });
    };

    var do_discovery = function (callback) {
        snmpDiscoverySnapshotData = {};
        var communitiesProvider = resourceLocator.snmpCommunitiesProvider;
        return communitiesProvider.get().then(function (communities) {
            var devicePromises = [];
            interfacesBinding.listFresh().forEach(function (device) {
                if (macToOidList[device.mac] === undefined) {
                    return;
                }
                my_console.info("Gathering information for %s", device.ip);
                var community = (communities[device.mac] && communities[device.mac].snmp_read_community) || DEFAULT_READ_COMMUNITY;
                var snmpCommand = commandBuilder.buildBulkGetCommand(macToOidList[device.mac], device.ip, community,
                    SNMP_DISCOVERY_TIMEOUT);
                snmpCommand.customTransformer = snmpCommon.type_invariant_transformer;

                devicePromises.push(deferredSnmpBulkGet(snmpCommand, device.ip, device.mac));
            });

            return q.all(devicePromises)
                .then(renderDiscoveryData)
                .then(function () {
                    unlock_discovery();
                    if (callback) {
                        callback(snmpDiscoverySnapshotData);
                    }
                });
        });
    };

    var discover = function (callback, errCallback) {
        if (!lock_discovery()) {
            my_console.warn("WARNING: SNMP Custom Discovery already in progress, ignoring the call");
            return q('BUSY');
        }

        return getOidListToDiscover().then(function () {
            return do_discovery(callback);
        }).then(function () {
            return 'FREE';
        }).catch(function (err, err2) {
            my_console.error('Error performing discovery: %s %s', err, err2);
            errCallback(err, err2);
        }).finally(function () {
            unlock_discovery();
            my_console.debug("Discovery unlocked");
        });

    };


    var getOidListToDiscover = function () {
        if (macToOidList === null) {
            my_console.debug('Requesting OID sensors to engine');
            return engineRequest.sendWithPromise('/oid-sensor', 'GET').then(function (response) {
                macToOidList = response;
                my_console.debug("Counting %s oid sensors", Object.keys(macToOidList).length);
            });
        } else {
            my_console.debug('Using OID sensors to inspect in memory');
            return q();
        }
    };


    /**
     * Send to engine all data gathered from devices with one HTTP post
     * @param discoveryData data gathered from devices
     */

    var sendDiscoveryDataToEngine = function (discoveryData) {
        var gotOIDS = Object.keys(discoveryData).length;
        if (gotOIDS === 0) {
            my_console.info("No OID data gatherted, sending nothing to engine");
        } else {
            my_console.info(" Records to send %d", gotOIDS);
            try {
                var resource = "/oid-sensor/sample";
                my_console.info(" Sending data to engine");

                engineRequest.sendWithPromise(resource, 'POST', discoveryData, null, null,
                    {doNotFollow204: true})
                    .catch(function (e) {
                        my_console.error("Error sending SNMP data to engine: %s", JSON.stringify(e));
                    });
            } catch (e) {
                my_console.error("Error preparing data to send %s", JSON.stringify(e));
            }
        }
    };

    return {
        discover: discover,
        invalidate: invalidate,
        sendDiscoveryDataToEngine: sendDiscoveryDataToEngine,
        isDiscoveryInProgress: function () {
            return mutex.isLocked(DISCOVERY_SEMAPHORE);
        },
        getCurrentSnapshotData: function () {
            return snmpDiscoverySnapshotData;
        }
    };
};
module.exports.factory = snmpDiscoveryFactory;
