/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz UK LLP
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 **/
/**
 * Created by Massimiliano Cuzzoli <massi@domotz.com> on 18/10/16.

 */

const DISCOVERY_SEMAPHORE = 'SNMP_INTERFACE_SEMAPHORE';
const OIDS_TO_DISCOVER = {
    '1.3.6.1.2.1.2.2.1.3': 'int',         // ifType
    '1.3.6.1.2.1.31.1.1.1.1': 'string',   // ifName
    '1.3.6.1.2.1.2.2.1.2': 'string',      // ifDescription
    '1.3.6.1.2.1.31.1.1.1.15': 'int',     // ifHighSpeed
    '1.3.6.1.2.1.31.1.1.1.18': 'string',  // ifAlias
    '1.3.6.1.2.1.2.2.1.7': 'int',         // ifAdminStatus
    '1.3.6.1.2.1.2.2.1.8': 'int',         // ifOperationalStatus
    '1.3.6.1.2.1.2.2.1.10': 'int',        // ifInOctets
    '1.3.6.1.2.1.2.2.1.16': 'int',        // ifOutOctets
    '1.3.6.1.2.1.2.2.1.13': 'int',        // ifInDiscards
    '1.3.6.1.2.1.2.2.1.19': 'int',        // ifOutDiscards
    '1.3.6.1.2.1.2.2.1.14': 'int',        // ifInErrors
    '1.3.6.1.2.1.2.2.1.20': 'int'         // ifOutErrors
};
const DEFAULT_READ_COMMUNITY = 'public';
const SNMP_SUBTREE_TIMEOUT = 20000;


var snmpDiscoveryFactory = function (resourceLocator, commandBuilder, responseRendering, snmpSubtreeFactory) {
    "use strict";
    var my_console = resourceLocator.log.decorateLogs("SNMP INTERFACE DISCOVERY - ");
    var snmpInterfacesNumberStorage = resourceLocator.snmpInterfacesNumberStorage;
    var communitiesProvider = resourceLocator.snmpCommunitiesProvider;
    var interfacesBinding = resourceLocator.interfacesBindingStorage;
    var mutex = resourceLocator.utils.mutex;
    var q = resourceLocator.q;
    var engineRequest = resourceLocator.engineRequest;
    var snmpDiscoverySnapshotData = {};

    var deferredSnmpSubtree = function (snmpCommand, host, mac) {
        var deferred = q.defer();

        snmpSubtreeFactory(snmpCommand, resourceLocator).execute(
            function (result) {
                my_console.debug('Received result for %s', host);
                result.host = host;
                result.mac = mac;
                result.timestamp = new Date().toISOString();
                deferred.resolve(result);
            }
        );

        return deferred.promise;
    };

    var lock_discovery = function () {
        return mutex.lock(DISCOVERY_SEMAPHORE, 1);
    };

    var unlock_discovery = function () {
        mutex.unlock(DISCOVERY_SEMAPHORE, 1);
    };

    var gatherInterfaceInfoForDevice = function (ipAddress, community, mac) {
        var snmpCommand = commandBuilder.buildSubtreeCommand(OIDS_TO_DISCOVER, ipAddress,
            community || DEFAULT_READ_COMMUNITY, SNMP_SUBTREE_TIMEOUT);
        my_console.info("Gathering information for %s (%s)", ipAddress, mac);
        return deferredSnmpSubtree(snmpCommand, ipAddress, mac);
    };

    var renderDiscoveryData = function (rawResultList) {
        for (var i = 0; i < rawResultList.length; i++) {
            var dataToRender = rawResultList[i];
            my_console.debug("Rendering discovery data for %s", dataToRender.host);
            snmpDiscoverySnapshotData[dataToRender.mac] = responseRendering(dataToRender);
        }
    };

    var discover = function (callback, errCallback) {

        if (!lock_discovery()) {
            my_console.warn("Already in progress, ignoring the call");
            return;
        }

        return communitiesProvider.get().then(function (communities) {
            var deferred_result_list = [];
            interfacesBinding.listFresh().forEach(function (device) {
                var interfaces = snmpInterfacesNumberStorage.fetch(device.mac);
                my_console.debug("Device %s has %s interfaces", device.mac, interfaces);
                if (interfaces > 0) {
                    var community = (communities[device.mac] && communities[device.mac].snmp_read_community) || DEFAULT_READ_COMMUNITY;
                    deferred_result_list.push(gatherInterfaceInfoForDevice(device.ip, community, device.mac));
                }
            });
            return q.all(deferred_result_list)
                .then(function (results) {
                    renderDiscoveryData(results);
                    unlock_discovery();
                    if (callback) {
                        callback(snmpDiscoverySnapshotData);
                    }
                })
                .catch(function (err) {
                    unlock_discovery();
                    my_console.error("Error on snmp information processing", err);
                    if (errCallback) {
                        errCallback(err);
                    }
                });

        });
    };

    var sendDiscoveryDataToEngine = function (discoveryData) {
        my_console.info("Number of records to send %d", Object.keys(discoveryData).length);
        try {
            var logError = function (e) {
                my_console.error("Error sending snmp data to Engine: %s", JSON.stringify(e));
            };
            for (var mac in discoveryData) {
                if (discoveryData.hasOwnProperty(mac) && discoveryData[mac] && discoveryData[mac].oids) {
                    my_console.info("Sending data for device MAC=%d", mac);
                    engineRequest.sendWithPromise('/subnet/ip/device/' + mac + '/interface/snmp-info', 'PUT',
                        discoveryData[mac]).catch(logError);
                }
            }
        } catch (e) {
            my_console.error("Error preparing data to send %s", JSON.stringify(e));
        }
    };

    return {
        discover: discover,
        sendDiscoveryDataToEngine: sendDiscoveryDataToEngine,
        isDiscoveryInProgress: function () {
            return mutex.isLocked(DISCOVERY_SEMAPHORE);
        },
        getCurrentSnapshotData: function () {
            return snmpDiscoverySnapshotData;
        }
    };
};


module.exports.factory = snmpDiscoveryFactory;
