/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created by Iacopo Papalini <iacopo@domotz.com> on 08/09/16.
 *
 * You need fing installed on the device in order to enable the device
 * network jamming
 */
const JAMMED_FILE_NAME = "jammed.txt";
const LOG_PREFIX = "FING_JAM_DEVICES - ";
const FING_LIB_PATH = "/usr/local/lib/fing";
const FING_BIN_DEFAULT_PATH = "/usr/local/lib/fing/fing.bin";
const FING_ENV_OPTIONS = {
    env: {
        'LD_LIBRARY_PATH': FING_LIB_PATH,
        'DYLD_LIBRARY_PATH': FING_LIB_PATH
    }
};

var path = require('path');

var jamDevicesFileManager = function (macListFile) {
    "use strict";
    var fs = require('fs');
    var previousMacAddresses = null;
    var internetPauseAvailable = null;
    var INTERNET_PAUSE_SUFFIX = ',internet-pause';

    function listChanged(macAddresses) {
        if (!previousMacAddresses) {
            return true;
        }
        if (macAddresses.length !== previousMacAddresses.length) {
            return true;
        }
        else {
            for (var i = macAddresses.length; i-- > 0;) {
                if (macAddresses[i] !== previousMacAddresses[i]) {
                    return true;
                }
            }
        }
        return false;
    }

    function writeFile(contents) {
        console.debug(LOG_PREFIX + 'Synchronously writing file %s', macListFile);
        try {
            fs.writeFileSync(macListFile, contents + '\n');
            console.debug(LOG_PREFIX + 'File written %s (%s)', macListFile, JSON.stringify(contents));
            return contents.length > 0;
        } catch (e) {
            console.error("Cannot write file %s, killing fing", macListFile);
            return false;
        }
    }

    function writeFileIfListChanged(macAddresses) {
        /**
         * Using the macAddresses list, writes its contents in the macListFile.
         * @returns boolean true if the fing process should run after the update
         */
        if (listChanged(macAddresses)) {
            console.info(LOG_PREFIX + "Jammed device list changed");
        } else {
            return macAddresses.length > 0;
        }
        previousMacAddresses = macAddresses;

        var contents = '';
        if (internetPauseAvailable && macAddresses.length > 0) {
            contents = macAddresses.join(INTERNET_PAUSE_SUFFIX + '\n') + INTERNET_PAUSE_SUFFIX;
        } else {
            contents = macAddresses.join('\n');
        }
        return writeFile(contents);
    }

    function setInternetPauseAvailable(value) {
        console.info("%s Setting internet pause availability as: %s", LOG_PREFIX, value);
        internetPauseAvailable = value;
    }

    function isInternetPauseAvailable() {
        /*
         * returns null if the information is not available yet,
         * true for versions >= 4.x, false otherwise
         *
         * */
        return internetPauseAvailable;
    }

    return {
        write: writeFileIfListChanged,
        setInternetPauseAvailable: setInternetPauseAvailable,
        isInternetPauseAvailable: isInternetPauseAvailable
    };
};


var fingProcessManager = function (fingCommand, macListFile, childProcess, procManager) {
    "use strict";
    var fingProcess = null;
    var fingParameters = ["--kickout", '-', '--kocfg', macListFile];

    function isFingRunning() {
        if (fingProcess !== null) {
            console.debug(LOG_PREFIX + "Fing process is running");
            return true;
        } else {
            console.debug(LOG_PREFIX + "Fing process not existing");
            return false;
        }
    }

    function assureFingIsRunning() {
        if (!isFingRunning()) {
            console.info(LOG_PREFIX + 'Starting fing');

            fingProcess = childProcess.spawn(fingCommand, fingParameters, {env: getEnrichedEnv()});

            fingProcess.on('close', function () {
                console.info("%s Fing process killed", LOG_PREFIX);
                fingProcess = null;
            });

            fingProcess.stderr.on('data', function (data) {
                console.warn(LOG_PREFIX + "ERR DATA  %s", data);
            });

            fingProcess.stdout.on('data', function (data) {
                console.verbose(LOG_PREFIX + "STDOUT DATA " + data);
            });

            fingProcess.on('error', function (err) {
                console.warn('%s Fing block call: unable to execute Fing: %s', LOG_PREFIX, JSON.stringify(err));
            });

        } else {
            console.debug(LOG_PREFIX + "Fing already running, not executed");
        }
    }

    function assureFingIsNotRunning() {
        if (isFingRunning()) {
            console.info(LOG_PREFIX + 'Killing fing');
            fingProcess.kill('SIGKILL');
            fingProcess = null;
        } else {
            console.info(LOG_PREFIX + "Fing not running, not killing it");
        }
    }

    function getEnrichedEnv() {
        var enrichedEnv = JSON.parse(JSON.stringify(process.env));
        enrichedEnv.LD_LIBRARY_PATH = (enrichedEnv.LD_LIBRARY_PATH ? enrichedEnv.LD_LIBRARY_PATH + ":" : "") + process.env.DOMOTZ_LIB_DIR + ":" + FING_ENV_OPTIONS.env.LD_LIBRARY_PATH;
        enrichedEnv.DYLD_LIBRARY_PATH = (enrichedEnv.DYLD_LIBRARY_PATH ? enrichedEnv.DYLD_LIBRARY_PATH + ":" : "") + process.env.DOMOTZ_LIB_DIR + ":" + FING_ENV_OPTIONS.env.LD_LIBRARY_PATH;
        return enrichedEnv;
    }

    function getFingVersion(callback) {
        getEnrichedEnv();
        procManager.exec(fingCommand + " --version", {env: getEnrichedEnv()}).then(function (stdout) {
            callback(null, stdout.trim());
        }).catch(function (error) {
            callback(error);
        });
    }

    return {
        assureFingIsRunning: assureFingIsRunning,
        assureFingIsNotRunning: assureFingIsNotRunning,
        getFingVersion: getFingVersion
    };
};


var fingWrapper = function (childProcess, engineRequest, procManager) {
    "use strict";
    var env = process.env;
    var macListFile = path.join((env.DOMOTZ_CONF_DIR), JAMMED_FILE_NAME);
    var fingCommand = (env.FING_BINARY_FULL_PATH || FING_BIN_DEFAULT_PATH);
    var fileWriter = jamDevicesFileManager(macListFile);
    var processManager = fingProcessManager(fingCommand, macListFile, childProcess, procManager);

    function handleServerResponse(data) {
        var shouldRun = fileWriter.write(data);
        if (shouldRun && data.length > 0) {
            processManager.assureFingIsRunning();
        }
        else {
            processManager.assureFingIsNotRunning();
        }
    }

    function updateMacAddresses() {
        console.info(LOG_PREFIX + "Updating mac addresses list");

        engineRequest.sendWithPromise('/device/jammed', 'GET').
            then(function (data) {
                handleServerResponse(data);
            }).
            catch(function (error) {
                console.error('%s Cannot update list, assuming no device to jam, %s', LOG_PREFIX, JSON.stringify(error));
                if (error.status.toString()[0] == '4') {
                    console.error("%s Client error %s, set jammed list as empty", LOG_PREFIX, error.status);
                    handleServerResponse([]);
                } else {
                    console.error("%s Server error %s, leave jammed list untouched", LOG_PREFIX, error.status);
                }
            });
    }

    return {
        updateMacAddresses: function () {
            if (fileWriter.isInternetPauseAvailable() === null) {
                processManager.getFingVersion(function (error, version) {
                    if (error) {
                        console.warn("%s Unable to get Fing version, Internet Pause feature will NOT be available %s, %s", LOG_PREFIX, error.message);
                    } else {
                        console.info("%s Fing version is %s", LOG_PREFIX, version);
                        if (parseInt(version.split('.')[0]) >= 4) {
                            fileWriter.setInternetPauseAvailable(true);
                        } else {
                            console.info("%s Fing version %s does not support Internet Pause", LOG_PREFIX, version);
                            fileWriter.setInternetPauseAvailable(false);
                        }
                    }
                    updateMacAddresses();
                });
            } else {
                updateMacAddresses();
            }
        },
        killDeviceJammer: function () {
            processManager.assureFingIsNotRunning();
        },
        getFingVersion: processManager.getFingVersion
    };
};

module.exports.fingWrapper = fingWrapper;

// For testing purposes only
module.exports._fileManager = jamDevicesFileManager;
module.exports._fingProcessManager = fingProcessManager;
