/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created by Andrea Azzara <a.azzara@domotz.com> on 23/08/2017.
 */

const MODULE = "AGENT_SETTINGS";

function SettingsManager(resourceLocator) {
    "use strict";
    var fileUtils = resourceLocator.utils.file;
    var myConsole = resourceLocator.log.decorateLogs(MODULE + ' ConfigurationFileManager - ');
    var fileName = process.env.DOMOTZ_CONF_DIR + '/agent_settings.json';

    function updateCustomInterfaces(refreshBindingsList) {
        myConsole.info("Updating custom settings: %s", JSON.stringify(resourceLocator.configuration.settings));

        var updateHostInfo = function (ifaces) {
            resourceLocator.discovery.host.loadCustomInterfaces(ifaces);
            if (refreshBindingsList) {
                resourceLocator.interfacesBindingStorage.invalidateList();
            }
        };


        if (!resourceLocator.configuration.settings.external_subnets && !resourceLocator.configuration.settings.external_hosts) {
            myConsole.info("No external subnets nor hosts found");
            updateHostInfo({});
            return;
        }

        var interfaces = {};

        if (resourceLocator.configuration.settings.external_subnets) {
            myConsole.info("Updating external subnets: %s", JSON.stringify(resourceLocator.configuration.settings.external_subnets));

            resourceLocator.configuration.settings.external_subnets.forEach(function (i) {
                interfaces['ext' + i.id] = {
                    'custom': true,
                    'netmask': i.subnet_mask,
                    'ip': i.address,
                    'mac': null
                };
            });
        } else {
            myConsole.info("No external subnets found");
        }


        if (resourceLocator.configuration.settings.external_hosts) {
            myConsole.info("Updating external hosts: %s", JSON.stringify(resourceLocator.configuration.settings.external_hosts));

            resourceLocator.configuration.settings.external_hosts.forEach(function (h) {
                interfaces['ext' + h.id] = {
                    'custom': true,
                    'netmask': 32,
                    'ip': h.ip_address,
                    'mac': null,
                    'hostname': h.hostname,
                    'discovery_hw_address': h.hw_address
                };
            });
        }
        else {
            myConsole.info("No external host found");
        }
        updateHostInfo(interfaces);
    }

    function loadSettings() {
        myConsole.info("Reading settings from cache: %s", fileName);
        var cachedSettings = fileUtils.readFile(fileName);
        if (cachedSettings) {
            resourceLocator.configuration.settings = JSON.parse(cachedSettings);
            myConsole.info("Found settings from cache: %s", JSON.stringify(resourceLocator.configuration.settings));
            updateCustomInterfaces();
        } else {
            myConsole.info("Settings not found in cache, requiring settings from server");
            fetchSettings().then(updateCustomInterfaces);
        }
    }

    function writeSettings(settings) {
        var fileContents = JSON.stringify(settings, null, '  ');
        myConsole.info("Writing agent settings: %s", fileName);
        fileUtils.writeFileSafe(fileName, fileContents);
    }

    function fetchSettings(refreshBindingsList) {
        var engineRequest = resourceLocator.engineRequest;

        return engineRequest.sendWithPromise('/settings', 'GET', {}, null, false, {
            maxAttempts: 3,
            retryDelay: 5000
        }).then(function (result) {
            myConsole.info('Updating settings with ' + JSON.stringify(result));
            resourceLocator.configuration.settings = result;
            writeSettings(result);
            updateCustomInterfaces(refreshBindingsList);
        }).catch(function (error) {
            myConsole.error(error);
        });
    }

    function isBroadcastDiscoveryDisabled() {
        if (resourceLocator.configuration.settings && resourceLocator.configuration.settings.broadcast_discovery !== undefined && resourceLocator.configuration.settings.broadcast_discovery === false) {
            return true;
        }
        return false;
    }

    function isLowThrottleMode() {
        if (resourceLocator.configuration.settings && resourceLocator.configuration.settings.http_call_throttle !== undefined && resourceLocator.configuration.settings.http_call_throttle === 'LOW') {
            return true;
        }
        return false;
    }

    function getEyesInterval() {
        if (resourceLocator.configuration.settings && resourceLocator.configuration.settings.eyes_hourly_frequency && resourceLocator.configuration.settings.eyes_hourly_frequency != 0) {
            return parseInt(120 / resourceLocator.configuration.settings.eyes_hourly_frequency);
        }
        return null;
    }

    return {
        loadSettings: loadSettings,
        writeSettings: writeSettings,
        fetchSettings: fetchSettings,
        isBroadcastDiscoveryDisabled: isBroadcastDiscoveryDisabled,
        isLowThrottleMode: isLowThrottleMode,
        getEyesInterval: getEyesInterval
    };

}

module.exports.settingsManager = SettingsManager;
