/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created by Iacopo Papalini <ipapalini@domotz.com> on 16/06/15.
 */

const MODULE = 'INIT';

// External Dependencies
var amqplib = require('amqplib');
var async = require('async');
var bodyParser = require('body-parser');
var childProcess = require('child_process');
var crypto = require('crypto');
var deepEqual = require('deep-equal');
var dgram = require('dgram');
var dhcpjs = require('../modules/dhcpjs');
var dns = require('dns');
var express = require('express');
var events = require('events');
var favicon = require('serve-favicon');
var fingHeartbeat = require('fing-heartbeat');
var fs = require('fs');
var http = require('http');
var ip = require('domotz-ip-address');
var lodash = require('lodash');
var moment = require('moment-timezone');
var natUpnp = require('domotz-nat-upnp');
var net = require('net');
var netSnmp = require('net-snmp');
var snmpNative = require('snmp-native');
var os = require('os');
var path = require('path');
var rootRequire = require('root-require');
var q = require('q');
var telnet = require('telnet-client');
var ssdp = require('domotz-node-ssdp');
var step = require('step');
var stringify = require('json-stable-stringify');
var url = require('url');
var util = require('util');
var xml2js = require('xml2js');
var xml2jsProcessor = require('xml2js/lib/processors.js');
var whois = require('node-whois');
var winston = require('winston');
var ws = require("ws");
var _ = require('underscore');

// Internal Dependencies (no need for further initialization)
var pkgJson = rootRequire('package.json');

var request = rootRequire('src/http/request');
var web = rootRequire('src/http/utils');

var command = rootRequire('src/utils/command');
var mutex = rootRequire('src/utils/mutex');
var time = rootRequire('src/utils/time');
var number = rootRequire('src/utils/number');
var random = require('../utils/random');
var modules = require('../utils/modules');
var dataStructure = rootRequire('src/utils/data_structure');
var scheduler = rootRequire('src/utils/scheduler');

var geoLocation = require('../network_tools/geolocation');
var commandHttpRequest = require('../commands/httpRequest');
var snmpCommon = rootRequire('src/snmp/common');
var snmpGetNative = rootRequire('src/commands/snmp/getNative');

function initFileManager() {
    console.debug("%s - Creating file manager", MODULE);
    return require('../persistence/file').factory(fs, q);
}

function getCacheUtils(procManager) {
    console.debug("%s - Creating cache manager", MODULE);
    return require('../persistence/cache').factory(procManager, fs);
}


function getHubConfigurator(request, file, web) {
    console.debug("%s - Loading HUB configuration", MODULE);
    return require('../setup/configuration').factory(q, request, file, web, path);
}

function initStatusManager(configuration) {
    console.debug("%s - Initializing status manager", MODULE);
    return require('../setup/status').factory(configuration);
}

function initLog(file, proc, configuration) {
    console.info("%s - Initializing Log", MODULE);
    return require('../setup/logger').factory(_, winston, util, path, file, proc, configuration);
}

function initIpWrapper(resourceLocator) {
    "use strict";
    return require('../net/main_network').ipWrapper(resourceLocator);
}

function initEmergencyNetworkRestore(resourceLocator) {
    "use strict";
    return require('../network_tools/emergencyNetworkRestore').factory(resourceLocator);
}

function initExternalProcessesManager(domotzApp) {
    /**
     * For now the algorithm is very simple.
     * If openwrt only 10 executors, otherwise 100.
     */
    var cpus = process.env.DPLATFORM === 'openwrt' ? 10 : 100;
    console.debug("%s - Creating external processes executor: %d cpus", MODULE, cpus);
    var Queue = require('../external_processes/queue');
    var cpu = require('../external_processes/cpu').factory(_, childProcess, Queue, cpus);
    return require('../external_processes/proc').factory(q, fs, childProcess, cpu, domotzApp);
}

function getHostInterfaces(procManager, fileManager, nmapXml2Js) {
    var networkInterfaces,
        customInterfaces = require('../discovery/host/custom').factory(_, fileManager);
    var netUtilsPath = path.join(process.env.DOMOTZ_BIN_DIR, 'domotz_network_utils');
    var v2 = fileManager.getFileStats(netUtilsPath);
    var platform = os.platform();
    if (!v2) {
        var nic = require('../discovery/host/nic').factory(util, platform);
        networkInterfaces = require('../discovery/host/interfaces').factory(q, procManager, nic);
    }
    console.debug("%s - Recovering host network interfaces with %s tools. Platform ==> %s", MODULE, v2 ? 'domotz' : 'local', platform);
    return require('../discovery/host').factory(procManager, ip, networkInterfaces, customInterfaces, nmapXml2Js, lodash, q, dns, fs);
}

function getAmqpListener(proc, status, configuration, eventHandler, resourceLocator) {
    var amqpHandler = require('../amqp/handler').factory(resourceLocator);
    var amqpConnection = require('../amqp/connection').factory(amqplib, util, configuration);
    var amqpConsumer = require('../amqp/consumer').factory(configuration, amqpHandler);
    var amqpChannel = require('../amqp/channel').factory(eventHandler);
    console.debug("%s - Initializing AMQP listener", MODULE);
    return require('../amqp/listener').factory(proc, status, amqpConnection, amqpConsumer, amqpChannel);
}

function initAvahiResolver(proc, engineRequest, devicesInfo) {
    console.debug("%s - Initializing avahi resolver", MODULE);
    var avahiCommands = require('../discovery/avahi/commands').factory();
    return require('../discovery/avahi/resolve').factory(engineRequest, devicesInfo, avahiCommands, proc);
}

function prepareNmapTranslator() {
    return require('../discovery/nmap/xml2js').factory(q, xml2js);
}

function setupPortDiscovery(resourceLocator, nmapCommands) {
    return require('../discovery/ports/scan').factory(resourceLocator, nmapCommands);
}

function setupTcpServicesMonitoring(proc, nmapXml2js, nmapCommands, interfacesBindingStorage, engineRequest, configuration) {
    var tcpServicesParser = require('../discovery/tcp_services/parser');
    return require('../discovery/tcp_services/tcp_monitor').factory(q, proc, nmapCommands, nmapXml2js, engineRequest, configuration, tcpServicesParser, interfacesBindingStorage);
}

function setupDiscoveryUtils(interfacesBindingStorage, devicesInfo) {
    return require('../discovery/utils').factory(interfacesBindingStorage, devicesInfo);
}

function setupProtocolsDiscoveryOrchestrator(domotzApp) {
    return require('../discovery/protocols/protocols_discovery_orchestrator').factory(domotzApp, domotzApp.discoverySenderFactory);
}

function setupBonjourDiscovery() {
    console.debug('%s - Starting Bonjour Server', MODULE);
    return require('../discovery/bonjour/bonjour_discovery').factory(mutex);
}

function setupUpnpDeviceDiscovery(resourceLocator) {
    var messageBuilder = require('../discovery/upnp/message').factory(resourceLocator);
    var responseParser = require('../discovery/upnp/parser').factory(resourceLocator, messageBuilder);
    var deferredRequestHandler = require('../discovery/upnp/deferred_request').factory(resourceLocator, messageBuilder);
    return require('../discovery/upnp/upnp_discovery').factory(resourceLocator, messageBuilder, responseParser, deferredRequestHandler);
}

function setupSnmpBasicDiscovery(resourceLocator) {
    var commandBuilder = require('../discovery/snmp/command_builder').factory(resourceLocator);
    var responseRendering = require('../discovery/snmp/response_processing').renderSnmpBasicData;
    var snmpGetterFactory = require('../commands/snmp/get.js').command;
    return require('../discovery/snmp/snmp_basic_discovery').factory(resourceLocator, commandBuilder, responseRendering, snmpGetterFactory);
}

function setupSnmpCustomDiscovery(resourceLocator) {
    var commandBuilder = require('../discovery/snmp/command_builder').factory(resourceLocator);
    var responseRendering = require('../discovery/snmp/response_processing').renderSnmpBasicData;
    var snmpGetterFactory = require('../commands/snmp/get.js').command;
    return require('../discovery/snmp/snmp_custom_discovery').factory(resourceLocator, commandBuilder,
        responseRendering, snmpGetterFactory);
}

function setupSnmpInterfaceDiscovery(resourceLocator) {
    var commandBuilder = require('../discovery/snmp/command_builder').factory(resourceLocator);
    var responseRendering = require('../discovery/snmp/response_processing').renderSnmpInterfaceData;
    var snmpSubtreeFactory = require('../commands/snmp/subtrees.js').command;

    return require('../discovery/snmp/snmp_interface_discovery').factory(resourceLocator, commandBuilder, responseRendering, snmpSubtreeFactory);
}

function setupDhcpDiscovery(engineRequest, resourceLocator) {
    console.debug('%s - Initialize DHCP server', MODULE);
    try {
        return require('../discovery/dhcp/dhcp_discovery').factory(dhcpjs, random, engineRequest, resourceLocator);
    } catch (e) {
        console.error('%s - Error running DHCP server, this discovery will be disabled', MODULE);
        return null;
    }
}

function setPeriodicalTasks(status, eventHandler, configuration) {
    return require('../events/cron').factory(time, status, eventHandler, configuration, random);
}

function initAPIClient(resourceLocator, endpoint) {
    console.info('%s - Creating "%s" api client.', MODULE, endpoint);
    return require('../api_client/engineRequest').factory(resourceLocator, endpoint);
}

function initEvents(events, domotzApp) {
    console.debug('%s - Creating Domotz Event Emitter', MODULE);
    return require('../events/eventHandler').factory(events, domotzApp);
}

function initDevicesCache(engineRequest, configuration) {
    console.debug('%s - Creating devices cache', MODULE);
    return require('../devices/info').factory(q, async, time, engineRequest, configuration);
}

function createWebapp(status, configuration, eventHandler, domotzApp) {
    console.debug('%s - Creating local web server', MODULE);
    var routes = require('../http/routes/index').create(express, status, configuration, domotzApp);
    return require('../http/app').createApp(fs, http, express, eventHandler, bodyParser, favicon, path, routes);
}

function createWsServer(domotzApp) {
    return require('../network_tools/wsServer').wsServer(domotzApp);
}

function setSignalHandlers(proc, log, status, event, domotzApp) {
    require('../setup/signals').setHandler(proc, log, status, event, domotzApp);
}

function initUserEntity(userRequest, registerRequest) {
    return require('../api_client/entities/user').factory(userRequest, registerRequest);
}

function initAgentEntity(web, host, configuration, userRequest, platform) {
    return require('../api_client/entities/agent').factory(random, url, web, host, configuration, userRequest, platform);
}

function initNotConfiguredAgentEntity(host, pkgJson, registerRequest) {
    return require('../api_client/entities/notConfiguredAgent').factory(host, pkgJson, registerRequest);
}

function initEventEntity(engineRequest) {
    return require('../api_client/entities/event').factory(engineRequest);
}

function initSocketListener() {
    var requestHandler = require('../http/socketHandler');
    return require('../http/socketListener').socketListener(http, fs, requestHandler);
}

function setupUpnpRegistry(deepEqual) {
    return require('../upnp/upnpRegistry').registry(deepEqual);
}

function getNetInfoModule(proc, file, hostInfo) {
    var parser = require('../network_tools/netinfo_utils');
    return require('../network_tools/netinfo').factory(os, path, http, dns, whois, proc, file, hostInfo, parser);
}

function initUpnpEventHandler() {
    return require('../upnp/upnpEventHandler');
}

function setupMtrTool(proc) {
    return require('../network_tools/mtr').factory(_, proc, dataStructure);
}

function initUdpListener(request, configuration) {
    return require('../udp/listener').factory(crypto, dgram, request, configuration);
}

function getUdpUtils() {
    return require('../udp/utils').factory(dgram);
}

function initSshSendFactory(sshTool) {
    console.debug('%s - Initializing ssh Send driver', MODULE);
    return require('../commands/ssh/sshSendFactory').factory(async, q, sshTool, _);
}

// TODO remove after LKB-184 worker is in production
function initControl4Driver(sshTool) {
    console.debug('%s - Initializing Control4 driver', MODULE);
    return require('../discovery/zigbee/control4').factory(async, q, sshTool, _, util);
}

function getSshTool(proc, file) {
    console.debug('%s - Initialize ssh interface', MODULE);
    return require('../network_tools/ssh').factory(proc, file);
}

function initJamDriver(engineRequest, procManager) {
    console.debug('%s - Initializing Jam Driver', MODULE);
    if (process.env.DPLATFORM === 'luxul' && (process.env.DARCHITECTURE === 'bcm53xx' || process.env.DARCHITECTURE === 'octeon')) {
        console.debug('%s - Using Luxul Jam Driver', MODULE);
        return require('../network_tools/luxulJamDevices').jamDriver(engineRequest, path, q);
    } else {
        console.debug('%s - Using Fing Jam Driver', MODULE);
        return require('../network_tools/fingJamDevices').fingWrapper(childProcess, engineRequest,procManager);
    }
}

function initStallWatcher(procManager) {
    console.debug('%s - Initializing stall watcher', MODULE);
    return require('../external_processes/stallWatcher').init(os, procManager, fs, path);
}

function setupNdtServersFinder(request, cache, configuration) {
    var distanceCalculator = require('../network_tools/earthDistance').distanceCalculator(configuration);
    return require('../network_tools/ndtServers').factory(q, cache, configuration, request, random, distanceCalculator);
}

function setupSpeedTestTool(proc, configuration, mlabSpeedTest) {
    return require('../network_tools/speedTest').factory(_, proc, dataStructure, number, configuration, mlabSpeedTest);
}

function setupMlabSpeedTestTool(proc, fileManager) {
    console.debug('%s - Initialize MLAB Speed Test', MODULE);
    return require('../network_tools/mlabSpeedTest').factory(q, path, proc, fileManager);
}

function initDiscoveryCoordinator(proc, hostInfo, eventHandler, nmapCommands, nmapXml2Js, interfacesBindingStorage, engineRequest, configuration, _, settingsManager, nmapBlackList, log) {
    "use strict";
    var ipDiscoveryParser = require('../discovery/ip_network/discovery_parser').factory(interfacesBindingStorage, hostInfo);
    var bindingDiscoveryFactory = require('../discovery/ip_network/bind_discovery').factory(q, proc, hostInfo, nmapCommands, nmapXml2Js, ipDiscoveryParser, settingsManager, log, configuration);
    var discoveryRetry = require('../discovery/ip_network/discovery_retry').factory(q, hostInfo, log, configuration, nmapBlackList, interfacesBindingStorage, async, lodash, bindingDiscoveryFactory, nmapCommands);
    var engineReportingStrategy = require('../discovery/ip_network/engine_report_strategy').engineIpDiscoveryReportStrategyFactory(q, engineRequest, interfacesBindingStorage, configuration, lodash, hostInfo, log, settingsManager);

    console.debug("%s - Creating device discovery coordinator", MODULE);
    return require('../discovery/ip_network/discovery_coordinator').factory(eventHandler, bindingDiscoveryFactory, interfacesBindingStorage, engineReportingStrategy, discoveryRetry, hostInfo, log);
}

function getNmapCommands(configuration, nmapBlackList) {
    return require('../discovery/nmap/commands').factory(ip, configuration, nmapBlackList);
}

function initBindingsCache(fileManager, engineRequest, hostInfo, resourceLocator) {
    var deviceMap = path.join(process.env.DOMOTZ_CONF_DIR, 'device_map.conf');
    console.debug("%s - Initializing discovery cache for < MAC : IP > bindings. File for custom mappping: %s", MODULE, deviceMap);
    return require('../discovery/ip_network/interfaces_binding').factory(q, util, fileManager,
        engineRequest,
        hostInfo,
        deviceMap,
        resourceLocator);
}

function setupOoklaServerFinder(domotzApp) {
    return require('../network_tools/ooklaServers').factory(domotzApp);
}

function setupConnectivityChecker(domotzApp) {
    return require('../network_tools/connectivity').factory(domotzApp);
}

function setupNetworkManager(procManager, fileManager, resourceLocator) {
    return require('../network_tools/networkManager').factory(procManager, fileManager, resourceLocator);
}

function initUpnpSubscriptionManager(domotzApp) {
    return require('../upnp/upnpSubscriptionsManager').factory(domotzApp);
}

function createUpnpServer(domotzApp) {
    console.debug('%s - Creating old UPNP server for power plug management', MODULE);
    return require('../upnp/upnpHttpServer').httpServer(domotzApp);
}

function setupSettingsManager(domotzApp) {
    return require('../setup/agentSettings').settingsManager(domotzApp);
}

function createPlatformUtils(fs) {
    return require('../utils/platform').factory(fs);
}


module.exports.create = function (customLogger, forcedStatus) {

    var domotzApp = {};

    // Npm Modules (for command or events that needs stuff at runtime)
    domotzApp.async = async;
    domotzApp.crypto = crypto;
    domotzApp.dhcpjs = dhcpjs;
    domotzApp.fs = fs;
    domotzApp.http = http;
    domotzApp.ip = ip;
    //TODO: replace underscore with lodash
    domotzApp.lodash = lodash;
    domotzApp.moment = moment;
    domotzApp.natUpnp = natUpnp;
    domotzApp.netSnmp = netSnmp;
    domotzApp.os = os;
    domotzApp.snmpNative = snmpNative;
    domotzApp.net = net;
    domotzApp.path = path;
    domotzApp.q = q;
    domotzApp.request = request;
    domotzApp.telnet = telnet;
    domotzApp.step = step;
    domotzApp.stringify = stringify;
    domotzApp.ssdp = ssdp;
    domotzApp.url = url;
    domotzApp.util = util;
    domotzApp.xml2js = xml2js;
    domotzApp.xml2jsProcessor = xml2jsProcessor;
    domotzApp.ws = ws;
    domotzApp._ = _;

    try {
        domotzApp.netPing = require('net-ping');
    } catch (err) {
        console.debug("%s - net-ping module not available", MODULE);
        domotzApp.netPing = null;
    }

    console.debug("%s - Imported external dependencies", MODULE);

    // Utils
    var fileManager = initFileManager();
    var procManager = initExternalProcessesManager(domotzApp);
    var cacheUtils = getCacheUtils(procManager);

    domotzApp.utils = {
        proc: procManager,
        childProcess: childProcess,
        random: random,
        file: fileManager,
        fingHeartbeat: fingHeartbeat,
        mutex: mutex,
        scheduler: scheduler,
        web: web,
        time: time,
        command: command,
        number: number,
        cache: cacheUtils,
        dataStructure: dataStructure,
        modules: modules,
        platform: createPlatformUtils(fs)
    };

    console.debug("%s - Imported utilities", MODULE);

    // Pawn Configuration
    var configurator = getHubConfigurator(request, fileManager, web);
    var configuration = configurator.loadConfiguration();
    domotzApp.configurator = configurator;
    domotzApp.configuration = configuration;

    // Status Manager
    var status = initStatusManager(configuration);
    domotzApp.status = status;
    if (forcedStatus) {
        domotzApp.status.set(forcedStatus);
    }

    // Log Setup
    var domotzLog = initLog(fileManager, procManager, configuration);
    var nmapXml2Js = prepareNmapTranslator(q, xml2js);
    domotzApp.log = customLogger || domotzLog;
    // Host Information
    var host = getHostInterfaces(procManager, fileManager, nmapXml2Js);
    domotzApp.discovery = {
        host: host
    };

    // API Clients. For OpenAPI limit the number of Sockets
    http.globalAgent.maxSockets = 100;
    domotzApp.httpOutCache = require('../api_client/httpOutCache.js').httpOutCache(domotzApp);

    var registerRequest = initAPIClient(domotzApp, 'register');
    var userRequest = initAPIClient(domotzApp, 'user');
    var engineRequest = (status.isConfigured() || status.isPassive()) ? initAPIClient(domotzApp, 'agent') : null;
    domotzApp.registerRequest = registerRequest;
    domotzApp.userRequest = userRequest;
    domotzApp.engineRequest = engineRequest;

    // Entities
    var eventEntity = initEventEntity(engineRequest);
    domotzApp.entities = {
        agent: initAgentEntity(web, host, configuration, userRequest, domotzApp.utils.platform),
        notConfiguredAgent: initNotConfiguredAgentEntity(host, pkgJson, registerRequest),
        event: eventEntity
    };

    http.globalAgent.maxSockets = Infinity;

    // Events Management
    var eventHandler = initEvents(events, domotzApp);
    domotzApp.event = eventHandler;

    setSignalHandlers(procManager, domotzLog, status, eventEntity, domotzApp);

    // Package Management
    domotzApp.pkg = {
        json: pkgJson
    };

    // NetworkTools
    domotzApp.networkTools = {
        geolocation: geoLocation
    };

    // Internal Web Server
    domotzApp.webapp = createWebapp(status, configuration, eventHandler, domotzApp);

    // Web Sockets server
    domotzApp.wsServer = createWsServer(domotzApp);

    domotzApp.distanceCalculator = require('../network_tools/earthDistance').distanceCalculator(configuration);

    // Periodic Tasks
    domotzApp.periodicalTasks = setPeriodicalTasks(status, eventHandler, configuration);

    domotzApp.utils.ipWrapper = initIpWrapper(domotzApp);
    domotzApp.emergencyNetworkRestore = initEmergencyNetworkRestore(domotzApp);

    // Write pid file since luxul hasn't got the forever process manager
    if (process.env.DPLATFORM === 'luxul' && process.env.LISTENER_PID_FILE) {
        try {
            fs.writeFileSync(process.env.LISTENER_PID_FILE, process.pid);
        } catch (e) {
            console.error('Error writing pid file');
        }
    }

    if (status.isConfigured() || status.isPassive()) {

        var devicesInfo = initDevicesCache(engineRequest, configuration);
        var interfacesBindingStorage = initBindingsCache(fileManager, engineRequest, host, domotzApp);

        var discoveryUtils = setupDiscoveryUtils(interfacesBindingStorage, devicesInfo);

        var nmapBlackList = require('../discovery/nmap/blackList').factory(fileManager, configuration, interfacesBindingStorage);
        var nmapCommands = getNmapCommands(configuration, nmapBlackList);
        domotzApp.nmapBlackList = nmapBlackList;
        var sshTool = getSshTool(procManager, fileManager, q);

        // Amqp manager
        domotzApp.amqpListener = getAmqpListener(procManager, status, configuration, eventHandler, domotzApp);

        //// Engine Clients
        domotzApp.heartbeatRequest = initAPIClient(domotzApp, 'agent-heartbeat');

        // IP Discovery - Bindings Cache
        domotzApp.interfacesBindingStorage = interfacesBindingStorage;

        // Device Utilities
        domotzApp.devices = {
            info: devicesInfo
        };
        domotzApp.onvif = require('../net/onvif/onvif');

        // SNMP Utilities
        domotzApp.snmpCommon = snmpCommon;
        domotzApp.snmpGetNative = snmpGetNative.command;
        domotzApp.snmpInterfacesNumberStorage = require('../discovery/snmp/port_number_storage').storage;
        domotzApp.snmpCommunitiesProvider = require('../discovery/snmp/communities').communitiesProviderFactory(domotzApp);

        domotzApp.settingsManager = setupSettingsManager(domotzApp);
        domotzApp.settingsManager.loadSettings();

        // Discovery Engine
        domotzApp.discoverySenderFactory = require('../discovery/protocols/discovery_sender').factory;

        domotzApp.discovery.utils = discoveryUtils;
        domotzApp.discovery.avahi = initAvahiResolver(procManager, engineRequest, devicesInfo);
        domotzApp.discovery.port = setupPortDiscovery(domotzApp, nmapXml2Js, interfacesBindingStorage, nmapCommands);
        domotzApp.discovery.tcp_services_monitoring = setupTcpServicesMonitoring(procManager, nmapXml2Js, nmapCommands, interfacesBindingStorage, engineRequest, configuration);
        domotzApp.discovery.ip_network = {
            discoveryCoordinator: initDiscoveryCoordinator(procManager, host, eventHandler, nmapCommands, nmapXml2Js, interfacesBindingStorage, engineRequest, configuration, _, domotzApp.settingsManager, nmapBlackList, domotzApp.log)
        };
        domotzApp.discovery.dhcp_discovery = setupDhcpDiscovery(engineRequest, domotzApp);
        domotzApp.discovery.bonjour_discovery = setupBonjourDiscovery(domotzApp);
        domotzApp.discovery.upnp_device_discovery = setupUpnpDeviceDiscovery(domotzApp);
        domotzApp.discovery.upnp_igd_discovery = require('../discovery/upnp/igd_discovery').factory(domotzApp, engineRequest);
        domotzApp.discovery.snmp_basic_discovery = setupSnmpBasicDiscovery(domotzApp);
        domotzApp.discovery.snmp_custom_discovery = setupSnmpCustomDiscovery(domotzApp);
        domotzApp.discovery.snmp_interface_discovery = setupSnmpInterfaceDiscovery(domotzApp);
        domotzApp.discovery.protocols_orchestrator = setupProtocolsDiscoveryOrchestrator(domotzApp);
        domotzApp.discovery.onvifDiscovery = require('../discovery/onvif/onvif_discovery').factory(domotzApp);

        // Socket Http Server
        domotzApp.socketListener = initSocketListener(eventHandler);

        // Net Tools
        domotzApp.networkTools.netInfo = getNetInfoModule(procManager, fileManager, host);
        domotzApp.networkTools.mtr = setupMtrTool(procManager);
        domotzApp.networkTools.ndtServers = setupNdtServersFinder(request, cacheUtils, configuration);
        domotzApp.networkTools.ooklaServers = setupOoklaServerFinder(domotzApp);
        var mlabSpeedTest = setupMlabSpeedTestTool(procManager, fileManager);
        domotzApp.networkTools.mlabSpeedTest = mlabSpeedTest;
        domotzApp.networkTools.speedTest = setupSpeedTestTool(procManager, configuration, mlabSpeedTest);
        domotzApp.networkTools.internetConnectionStatus = setupConnectivityChecker(domotzApp);
        domotzApp.networkTools.ssh = sshTool;
        domotzApp.networkTools.networkManager = setupNetworkManager(domotzApp);
        domotzApp.networkTools.nmapCommands = nmapCommands;
        domotzApp.networkTools.rtdProcessor = require('../network_tools/rtd').rtdProcessor(domotzApp);
        domotzApp.networkTools.rtd = require('../network_tools/rtd').rtdScan(domotzApp);
        domotzApp.networkTools.rtd.loadFromFS();

        // SSH
        domotzApp.sshSendFactory = initSshSendFactory(sshTool);

        // Http command interface
        domotzApp.commandHttpRequest = commandHttpRequest;

        // Upnp Server
        domotzApp.upnp = {
            registry: setupUpnpRegistry(deepEqual),
            eventHandler: initUpnpEventHandler(),
            subscriptionsManager: initUpnpSubscriptionManager(domotzApp),
            httpServer: createUpnpServer(domotzApp)
        };

        // Udp Listener
        domotzApp.udp = {
            utils: getUdpUtils(),
            listener: initUdpListener(request, configuration)
        };

        // TODO remove after LKB-184 worker is in production
        // Control4
        domotzApp.control4Driver = initControl4Driver(sshTool);

        // Jam Driver
        domotzApp.jamDriver = initJamDriver(engineRequest, procManager);

        // Self check
        domotzApp.stallWatcher = initStallWatcher(procManager);

        // Local modules (e.g. Fing, Mlab)
        domotzApp.availableModules = {};
    } else {
        console.warn("%s - Hub is not configured (invalid or absent conf file). Running in restricted mode", MODULE);
    }

    var i = 0;
    for (var property in domotzApp) {
        if (domotzApp.hasOwnProperty(property)) {
            console.verbose('%s -   --> %s;', MODULE, property);
            i++;
        }
    }
    console.info('%s - Finish Initialization of Domotz Application. %d modules loaded', MODULE, i);
    return domotzApp;
};
